UNIT jcr6Filter;

INTERFACE

USES 
	MemTypes, Quickdraw, OSUtils, Packages, FilterInterface, Resources, Memory,
	Errors;
	

PROCEDURE EntryPoint (selector : INTEGER;
					  stuff : FilterRecordPtr;
					  VAR data : Handle;
					  VAR result : INTEGER);

IMPLEMENTATION

CONST kMaxCWidth = 4096; {support 4K wide color images}
	  kMaxWidth = 3*kMaxCWidth;	{support 12K width B/W images}
	  kFSize = 5; {# lines to examine at once}

TYPE 
	TParamRec = RECORD
					{parameters that need to be set up for this
					 filter, if any}
					selection : integer; {4=median,5=max,6=min,7=range
										  8=top hat+, 9=top hat-}
					isOctagonal : Boolean;
				END;
	TParamPtr = ^TParamRec;
	TParamH   = ^TParamPtr;
	
	unsignedbyte = 0..255;
	
	C3LineType = Packed Array [0..kMaxWidth-1] of unsignedbyte;
		{this is for a line that is rgbrgbrgbrgb packed bytes}
	C3LinePtr = ^C3LineType;
	
	SArray=Array[1..25] of integer; {for sorting}

				  	
{There are NO GLOBAL VARIABLES unless you want to create an A5 world.
 All of our globals are part of our Handle - storage record, or the
 data storage handle}
 
PROCEDURE Dispatch   (selector : INTEGER;
					  stuff : FilterRecordPtr;
					  VAR data : Handle;
					  VAR result : INTEGER); FORWARD;
					  

PROCEDURE EntryPoint (selector : INTEGER;
					  stuff : FilterRecordPtr;
					  VAR data : Handle;
					  VAR result : INTEGER);

BEGIN
	Dispatch(selector, stuff, data, result);
END; {EntryPoint}

{assorted INLINE calls}

FUNCTION DoTestAbort(codeAddress : ProcPtr) : Boolean;  {true = user wants to stop}
	INLINE $205F,$4E90;

PROCEDURE DoUpdateProgress (done, total : Longint; codeAddress : ProcPtr);
	INLINE $205F,$4E90; 
		{get ProcPtr off stack,
		 jump to address with normal params on stack}


{utility routines}

PROCEDURE CenterDialog(atDLOG : handle);
CONST
	menuHeight = 20;
VAR
	r : Rect;
	atRect : ^Rect;
	ScreenWidth : integer;
	ScreenHeight : integer;
	ScreenPixMap: PixMapHandle;
	
	PROCEDURE Getscreen;
	CONST
		ROM85Loc = $28E;
	VAR
		WordPtr: ^integer;
		ColorQDExists: boolean;
		MainDevice: GDHandle;
	BEGIN
		WordPtr := Pointer(ROM85Loc);
		ColorQDExists := BAND(WordPtr^, $C000) = 0;
		MainDevice := GetMainDevice;
		ScreenPixMap := MainDevice^^.gdPMap;
		{ScreenRowBytes := BitAnd(ScreenPixMap^^.rowBytes, $1fff);}
		{ScreenBase := ScreenPixMap^^.baseAddr;}
	END; {GetScreen}

BEGIN
	IF atDLOG=NIL THEN EXIT(CenterDialog);
	GetScreen; {set up ScreenPixMap}
	WITH ScreenPixMap^^.bounds DO
		BEGIN
			ScreenWidth:=right;
			ScreenHeight:=bottom;
		END;{with}
	atRect:=pointer(atDLOG^);
	r:=atRect^;
	
	WITH r DO
		OffsetRect(r,-left,-top);
		
	WITH r DO
		OffsetRect(r,(ScreenWidth-right) div 2,
					 (ScreenHeight-bottom-menuHeight) div 3 + menuheight);
	atRect^:=r;
END;{centerdialog}

PROCEDURE DoAbout;
CONST 
	aboutDITL = 16000;
VAR
	mylog : dialogptr;
	item : integer;
	dt : handle;
BEGIN
	dt:=GetResource('DLOG',aboutDITL);
	IF dt=NIL THEN
		BEGIN
			Sysbeep(1);
			EXIT(DoAbout);
		END;
		
	{Never Never Never mess with a NIL handle}
	HNoPurge(dt);
	CenterDialog(dt);
	mylog:=GetNewDialog(aboutDITL, NIL, WindowPtr(-1));
	
	ModalDialog(NIL,item); {if the user clicks on anything, we're done}
	DisposDialog(mylog);
	HPurge(handle(dt)); {this could be dangerous}
END;{DoAbout}

PROCEDURE OutlineButton(mylog : dialogptr; item : integer);
VAR itembox : rect;
	itemhandle : handle;
	itemtype : integer;
	tport : grafptr;
BEGIN
	getport(tport);
	setport(grafptr(mylog));
	GetDItem(mylog,item,itemType,itemHandle,itembox);
	PenNormal;
	PenSize(3,3);
	InsetRect(itembox,-4,-4);
	FrameRoundRect(itembox,16,16);
	PenSize(1,1);
	setport(tport);
END;{OutlineButton}


FUNCTION GetDString(TheDialog:DialogPtr; item:integer):str255;
VAR
	ItemType:integer;
	ItemBox:rect;
	ItemHdl:handle;
	astr:str255;
BEGIN
	GetDItem (TheDialog,item,ItemType,ItemHdl,ItemBox);
	GetIText(ItemHdl,astr);
	GetDString:=astr;
END;{getdstring}

PROCEDURE SetDString(TheDialog:DialogPtr; item:integer; astr:str255);
VAR
	ItemType:integer;
	ItemBox:rect;
	ItemHdl:handle;
BEGIN
	GetDItem (TheDialog,item,ItemType,ItemHdl,ItemBox);
	SetIText(ItemHdl,astr)
END;{setdstring}

FUNCTION GetDNum(TheDialog:DialogPtr; item:integer):LongInt;
VAR
	ItemType:integer;
	ItemBox:rect;
	ItemHdl:handle;
	astr:str255;
	n:LongInt;
BEGIN
	GetDItem (TheDialog,item,ItemType,ItemHdl,ItemBox);
	GetIText(ItemHdl,astr);
	StringToNum(astr,n);
	GetDNum:=n;
END;{getdnum}

PROCEDURE SetDNum(TheDialog:DialogPtr; item:integer; n:longint);
VAR
	ItemType:integer;
	ItemBox:rect;
	ItemHdl:handle;
	astr:str255;
BEGIN
	GetDItem (TheDialog,item,ItemType,ItemHdl,ItemBox);
	NumToString(n,astr);
	SetIText(ItemHdl,astr)
END;{setdnum}

PROCEDURE SetDialogItem(TheDialog:DialogPtr; item,value:integer);
VAR
	ItemType:integer;
	ItemBox:rect;
	ItemHdl:handle;
BEGIN
	GetDItem (TheDialog,item,ItemType,ItemHdl,ItemBox);
	SetCtlValue(ControlHandle(ItemHdl),value)
END; {setdialogitem}


PROCEDURE Dispatch; {prototype declared above}
VAR gresult : oserr; {to be passed out}
	isTopEdge : boolean;
	isBottomEdge : boolean;

	FUNCTION CommandPeriod : Boolean;
	BEGIN
		CommandPeriod:=DoTestAbort(stuff^.abortProc);
	END;{CommandPeriod}
	
	PROCEDURE ShowProgress (done,total : longint);
	BEGIN
		DoUpdateProgress(done,total,stuff^.progressProc);
	END;{ShowProgress}

	PROCEDURE DoParameters;
		{ask user for plug-in's parameters.  The image size is undefined
		 at this point.  this may not be called each time that the
		 plug-in is executed}
	CONST
		ParamDITL = 16001;
	VAR r : rect;
		tport : grafptr;
		dt : Handle;
		item : integer;
		mylog : DialogPtr;
		myselection : integer; {4=median,5=max,6=min,7=range
								8=top hat+, 9=top hat-}
		DoOctagonal : Boolean;
	BEGIN
		InitCursor;
		
		IF stuff^.parameters=NIL THEN
			WITH stuff^ DO
				BEGIN
					{allocate space for parameters to be stored}
					parameters:=NewHandleClear(sizeof(TParamRec));
					IF parameters=NIL THEN
						BEGIN
							gResult:=memFullErr; {-108}
							EXIT(DoParameters);
						END;
					TParamH(parameters)^^.selection:=0; {set up default
													 parameter values}
					TParamH(parameters)^^.IsOctagonal:=false;		 
				END; {if,with - parameters=NIL}
				
		WITH stuff^ DO
			BEGIN
				TParamH(parameters)^^.selection:=4; {set up default
												 parameter values}
				TParamH(parameters)^^.IsOctagonal:=true;
				dt:=GetResource('DLOG',ParamDITL);
				if dt=NIL then
					BEGIN
						sysbeep(1);
						EXIT(DoParameters);
					END;
				{never mess with a NIL handle}
				HNoPurge(dt);
				CenterDialog(dt);
				mylog:=GetNewDialog(ParamDITL,NIL,WindowPtr(-1));
				
				mySelection:=TparamH(parameters)^^.selection;
				DoOctagonal:=TParamH(parameters)^^.IsOctagonal;
				SetDialogItem(mylog,mySelection,1); {radio button}
				SetDialogItem(mylog,3,ord(DoOctagonal)); {check box}
				item:=0;
				REPEAT
					ModalDialog(NIL,item);
					CASE item of
					1:;{okay}
					2:;{cancel}
					3:  BEGIN {octagonal check box}
							DoOctagonal:=NOT DoOctagonal;
							SetDialogItem(mylog,3,ord(DoOctagonal));
						END;
					4..9:{median,range,max,min radios}
						if item<>mySelection then
							BEGIN
								SetDialogItem(mylog,mySelection,0);
								mySelection:=item;
								SetDialogItem(mylog,item,1);
							END;
					END;{case}
				UNTIL (item=1) or (item=2);
				
				IF item=2 THEN gResult:=1 {guy hit cancel}
				ELSE
					BEGIN
						TParamH(parameters)^^.selection:=myselection;
						TParamH(parameters)^^.IsOctagonal:=DoOctagonal;
					END;{else good result}
				
				DisposDialog(mylog);
				hPurge(handle(dt));
			END;{with}
	END; {DoParameters}
	
	PROCEDURE DoPrepare;
	{if we require large amounts of memory, set the stuff^.bufferSpace
	 field to indicate how much (bytes)}
	BEGIN
		WITH stuff^ DO
			BEGIN
				{if we were setting up a large matrix, look-up tables, etc}
				stuff^.bufferSpace:=0;
			END; {with}
	END;{DoPrepare}
	
	PROCEDURE DoStart;
	{request pointer to first part of image to be filtered}
	BEGIN
		WITH stuff^ DO
			BEGIN
				{select top line, assume bottom is non-inclusive}
				inrect:=filterrect;
				inrect.bottom:=inRect.top+1; {want up to kFSize lines, first time
											  get only 1}
				isTopEdge:=true;
				isBottomEdge:=false;
				
				{select all planes}
				inLoPlane:=0;
				inHiPlane:=planes-1;
				
				outrect:=inRect;
				outrect.bottom:=outRect.top+1; {only write out 1 line}
				
				{output planes}
				outLoPlane:=0;
				outHiPlane:=planes-1;
			END; {with}
	END;{DoStart}
	
	PROCEDURE DoContinue;
	{filters the area requested and requests the next area}
	VAR x,			  	{x = horizontal position along line}
		zh : integer;	{local looping variable}
	
		S1Ptr,S2Ptr,	
		S3Ptr,S4Ptr,
		S5Ptr,			{S1Ptr..S5Ptr = pointers to start of each of 5 lines}
		T1Ptr : C3LinePtr;	{pointer to start of output line}
		
		width : integer;	{# pixels wide the line is}
		
		SortArray,LocalArea : SArray; {array[1..25] of integer for 5x5 neighborhood}
		SortSize : integer; {number to sort = 5,21,25 depending on n'borhood}
		min,max : integer;	{computed min & max for each pixel}
		
		glines : integer;	{# lines in current rect - usually 5, but could be as small as 1}
		
		myselection : integer;	{4=median, 5=max, 6=min, 7=range,
								 8=variance, 9=std. dev., 10=top hat}
		DoOctagonal : Boolean;
		sum : Longint;			{add up sum of squares of pixel differences}
		temp,cell : integer;	{hold pixel values for arithmetic}
		i,k : integer;
		
		
		PROCEDURE ShellSort(VAR A:SArray; N:Integer);
		VAR
			Done:Boolean;
			Jump,i,j,Hold:integer;
		BEGIN
			Jump:=N;
			While Jump>1 DO
				BEGIN
					Jump:=Jump DIV 2;
					Repeat
						Done:=True;
						For j:=1 to n-Jump DO
							BEGIN
								i:=j+Jump;
								IF A[j]>A[i] THEN
									BEGIN
										Hold:=A[j];
										A[j]:=A[i];
										A[i]:=Hold;
										Done:=False;
									END;
							END;
					Until Done;
				END;
		END;{ShellSort}
		
		FUNCTION IntegerSqrt(N:Longint):Longint;
		{perfect square is the sum of (root) sequential odd integers}
		VAR
			total,step,root:Longint;
		BEGIN
			total:=-1;root:=0;step:=-1;
			Repeat
				root:=root+1;
				step:=step+2;
				total:=total+step;
			Until (total>=N);
			IntegerSqrt:=root;
		END;{IntegerSqrt}
	
	BEGIN
		WITH stuff^ DO
			BEGIN
				myselection:=TParamH(parameters)^^.selection;
				DoOctagonal:=TParamH(parameters)^^.IsOctagonal;
				{check for cancel}
				IF CommandPeriod THEN
					BEGIN
						gResult:=1; {cancel}
						EXIT(DoContinue);
					END;
				
				S1Ptr:=pointer(inData);
				T1Ptr:=pointer(outData);

				glines:=inRect.bottom-inRect.top;
				CASE gLines OF
				1:  BEGIN
						S2Ptr:=S1Ptr;
						S3Ptr:=S1Ptr;
						S4Ptr:=S1Ptr;
						S5Ptr:=S1Ptr;
					END;
				2:	BEGIN
						S2Ptr:=pointer(ord4(S1Ptr)+inRowBytes);
						S3Ptr:=S2Ptr;
						S4Ptr:=S1Ptr;
						S5Ptr:=S1Ptr;
					END;
				3: BEGIN
						S2Ptr:=pointer(ord4(S1Ptr)+inRowBytes);
						S3Ptr:=pointer(ord4(S2Ptr)+inRowBytes);
						S4Ptr:=S1Ptr;
						S5Ptr:=S3Ptr;
				   END;
				4: BEGIN
						S2Ptr:=pointer(ord4(S1Ptr)+inRowBytes);
						S3Ptr:=pointer(ord4(S2Ptr)+inRowBytes);
						S4Ptr:=pointer(ord4(S3Ptr)+inRowBytes);
						S5Ptr:=S2Ptr;
				   END;
				5: BEGIN
						S2Ptr:=pointer(ord4(S1Ptr)+inRowBytes);
						S3Ptr:=pointer(ord4(S2Ptr)+inRowBytes);
						S4Ptr:=pointer(ord4(S3Ptr)+inRowBytes);
						S5Ptr:=pointer(ord4(S4Ptr)+inRowBytes);
				   END;
				END; {case}
				
				{process this portion}
				width:=inRect.right-inRect.left;
				
				IF planes<>3 THEN {B/W, or something else}
					BEGIN {note: myselection = 410; doOctagonal=true/false}
						width:=width*planes; {if planes<>1 and <>3}
						
						LocalArea[ 3]:=S1Ptr^[0];
						LocalArea[ 8]:=S2Ptr^[0];
						LocalArea[13]:=S3Ptr^[0];
						LocalArea[18]:=S4Ptr^[0];
						LocalArea[23]:=S5Ptr^[0];
						
						LocalArea[ 4]:=S1Ptr^[1];
						LocalArea[ 9]:=S2Ptr^[1];
						LocalArea[14]:=S3Ptr^[1];
						LocalArea[19]:=S4Ptr^[1];
						LocalArea[24]:=S5Ptr^[1];

						LocalArea[ 5]:=S1Ptr^[2];
						LocalArea[10]:=S2Ptr^[2];
						LocalArea[15]:=S3Ptr^[2];
						LocalArea[20]:=S4Ptr^[2];
						LocalArea[25]:=S5Ptr^[2];
						
						{col2 = col4}
						LocalArea[ 2]:=S1Ptr^[1];
						LocalArea[ 7]:=S2Ptr^[1];
						LocalArea[12]:=S3Ptr^[1];
						LocalArea[17]:=S4Ptr^[1];
						LocalArea[22]:=S5Ptr^[1];
						
						{col1 = col5}
						LocalArea[ 1]:=S1Ptr^[2];
						LocalArea[ 6]:=S2Ptr^[2];
						LocalArea[11]:=S3Ptr^[2];
						LocalArea[16]:=S4Ptr^[2];
						LocalArea[21]:=S5Ptr^[2];
						
						SortArray:=LocalArea;
						SortSize:=25;
						IF DoOctagonal THEN
							BEGIN {replace corners with values past 21}
								SortArray[1] :=SortArray[24];
								SortArray[5] :=SortArray[23];
								SortArray[21]:=SortArray[22];
								SortSize:=21;
							END;
						
						(*
						{alternative to sorting}
						IF (myselection=6) or (myselection=7) {not max only} THEN
							BEGIN
								min:=LocalArea[1];
								FOR zh:=2 TO SortSize DO
									IF min>LocalArea[zh] THEN min:=LocalArea[zh];
							END;
						
						IF (myselection=5) or (myselection=7) {not min only} THEN
							BEGIN
								max:=LocalArea[1];
								FOR zh:=2 TO SortSize DO
									IF max<LocalArea[zh] THEN max:=LocalArea[zh];
							END;
							
						CASE myselection OF
							5:T1Ptr^[0]:=max; 		{maximum}
							6:T1Ptr^[0]:=min;		{minimum}
							7:T1Ptr^[0]:=max-min; 	{range}
							END; {case}
						*)
						
						CASE myselection OF
							4:{median}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[0]:=SortArray[1+(SortSize div 2)];
								END;
							5:{Maximum}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[0]:=SortArray[SortSize];
								END;
							6:{Minimum}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[0]:=SortArray[1];
								END;
							7:{Range}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[0]:=SortArray[SortSize]-SortArray[1];
								END;
							(*
							{Variance and Std. Dev.}
								BEGIN
									Cell:=SortArray[13];{central pixel}
									Sum:=0;
									For i:=1 to SortSize DO
										BEGIN
											temp:=SortArray[i]-Cell;
											Sum:=Sum+temp*temp;
										END;
									Sum:=Sum div SortSize;
									if (myselection=9)
										THEN
											T1Ptr^[0]:=IntegerSqrt(Sum)
										ELSE
											T1Ptr^[0]:=sum div 256;
								END;
							*)
							8:{Top Hat+}
								BEGIN
									ShellSort(SortArray,SortSize);
									cell:=SortArray[SortSize];
									SortArray[1]:=LocalArea[8]; {pull out center}
									SortArray[2]:=LocalArea[12];
									SortArray[3]:=LocalArea[13];
									SortArray[4]:=LocalArea[14];
									SortArray[5]:=LocalArea[18];
									SortSize:=5;
									ShellSort(SortArray,SortSize);
									T1Ptr^[0]:=cell-SortArray[SortSize];
								END;
							9:{Top Hat-}
								BEGIN
									ShellSort(SortArray,SortSize);
									cell:=SortArray[1];
									SortArray[1]:=LocalArea[8]; {pull out center}
									SortArray[2]:=LocalArea[12];
									SortArray[3]:=LocalArea[13];
									SortArray[4]:=LocalArea[14];
									SortArray[5]:=LocalArea[18];
									SortSize:=5;
									ShellSort(SortArray,SortSize);
									T1Ptr^[0]:=SortArray[1]-cell;
								END;
							END; {CASE}
						
						FOR x:=1 TO width-3 DO {last 2 columns are interesting}
							BEGIN
								{rotate}
								FOR zh:=  1 TO  4 DO LocalArea[zh]:=LocalArea[zh+1];
												   	 LocalArea[ 5]:=S1Ptr^[x+2];
								FOR zh:=  6 TO  9 DO LocalArea[zh]:=LocalArea[zh+1];
												   	 LocalArea[10]:=S2Ptr^[x+2];
								FOR zh:= 11 TO 14 DO LocalArea[zh]:=LocalArea[zh+1];
													 LocalArea[15]:=S3Ptr^[x+2];
								FOR zh:= 16 TO 19 DO LocalArea[zh]:=LocalArea[zh+1];
													 LocalArea[20]:=S4Ptr^[x+2];
								FOR zh:= 21 TO 24 DO LocalArea[zh]:=LocalArea[zh+1];
													 LocalArea[25]:=S5Ptr^[x+2];
								SortArray:=LocalArea;
								SortSize:=25;
								IF DoOctagonal THEN
									BEGIN {replace corners with values past 21}
										SortArray[1] :=SortArray[24];
										SortArray[5] :=SortArray[23];
										SortArray[21]:=SortArray[22];
										SortSize:=21;
									END;
								
								(*
								{alternative to sorting}
								IF (myselection=6) or (myselection=7) {not max only} THEN
									BEGIN
										min:=LocalArea[1];
										FOR zh:=2 TO SortSize DO
											IF min>LocalArea[zh] THEN min:=LocalArea[zh];
									END;
								
								IF (myselection=5) or (myselection=7) {not min only} THEN
									BEGIN
										max:=LocalArea[1];
										FOR zh:=2 TO SortSize DO
											IF max<LocalArea[zh] THEN max:=LocalArea[zh];
									END;
									
								CASE myselection OF
									5:T1Ptr^[x]:=max; 		{maximum}
									6:T1Ptr^[x]:=min;		{minimum}
									7:T1Ptr^[x]:=max-min; 	{range}
									END; {case}
								*)
								
								CASE myselection OF
									4:{median}
										BEGIN
											ShellSort(SortArray,SortSize);
											T1Ptr^[x]:=SortArray[1+(SortSize div 2)];
										END;
									5:{Maximum}
										BEGIN
											ShellSort(SortArray,SortSize);
											T1Ptr^[x]:=SortArray[SortSize];
										END;
									6:{Minimum}
										BEGIN
											ShellSort(SortArray,SortSize);
											T1Ptr^[x]:=SortArray[1];
										END;
									7:{Range}
										BEGIN
											ShellSort(SortArray,SortSize);
											T1Ptr^[x]:=SortArray[SortSize]-SortArray[1];
										END;
									(*
									{Variance and Std. Dev.}
										BEGIN
											Cell:=SortArray[13];
											Sum:=0;
											For i:=1 to SortSize DO
												BEGIN
													temp:=SortArray[i]-Cell;
													Sum:=Sum+temp*temp;
												END;
											Sum:=Sum div SortSize;
											if (myselection=9)
												THEN
													T1Ptr^[x]:=IntegerSqrt(Sum) 
												ELSE
													T1Ptr^[x]:=sum div 256;
										END;
									*)
									8:{Top Hat+}
										BEGIN
											ShellSort(SortArray,SortSize);
											cell:=SortArray[SortSize];
											SortArray[1]:=LocalArea[8]; {pull out center}
											SortArray[2]:=LocalArea[12];
											SortArray[3]:=LocalArea[13];
											SortArray[4]:=LocalArea[14];
											SortArray[5]:=LocalArea[18];
											SortSize:=5;
											ShellSort(SortArray,SortSize);
											T1Ptr^[x]:=cell-SortArray[SortSize];
										END;
									9:{Top Hat-}
										BEGIN
											ShellSort(SortArray,SortSize);
											cell:=SortArray[1];
											SortArray[1]:=LocalArea[8]; {pull out center}
											SortArray[2]:=LocalArea[12];
											SortArray[3]:=LocalArea[13];
											SortArray[4]:=LocalArea[14];
											SortArray[5]:=LocalArea[18];
											SortSize:=5;
											ShellSort(SortArray,SortSize);
											T1Ptr^[x]:=SortArray[1]-cell;
										END;
									END; {CASE}
								
							END;{for}
							
						{x:=width-2;}
						
						{duplicate instead of rotate, col1 = col4}
						LocalArea[ 1]:=LocalArea[ 4];
						LocalArea[ 6]:=LocalArea[ 9];
						LocalArea[11]:=LocalArea[14];
						LocalArea[16]:=LocalArea[19];
						LocalArea[21]:=LocalArea[24];
						
						SortArray:=LocalArea;
						SortSize:=25;
						IF DoOctagonal THEN
							BEGIN {replace corners with values past 21}
								SortArray[1] :=SortArray[24];
								SortArray[5] :=SortArray[23];
								SortArray[21]:=SortArray[22];
								SortSize:=21;
							END;
						
						(*
						{alternative to sorting}
						IF (myselection=6) or (myselection=7) {not max only} THEN
							BEGIN
								min:=LocalArea[1];
								FOR zh:=2 TO SortSize DO
									IF min>LocalArea[zh] THEN min:=LocalArea[zh];
							END;
						
						IF (myselection=5) or (myselection=7) {not min only} THEN
							BEGIN
								max:=LocalArea[1];
								FOR zh:=2 TO SortSize DO
									IF max<LocalArea[zh] THEN max:=LocalArea[zh];
							END;
							
						CASE myselection OF
							5:T1Ptr^[width-2]:=max; 		{maximum}
							6:T1Ptr^[width-2]:=min;			{minimum}
							7:T1Ptr^[width-2]:=max-min; 	{range}
							END; {case}
						*)
						
						CASE myselection OF
							4:{median}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-2]:=SortArray[1+(SortSize div 2)];
								END;
							5:{Maximum}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-2]:=SortArray[SortSize];
								END;
							6:{Minimum}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-2]:=SortArray[1];
								END;
							7:{Range}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-2]:=SortArray[SortSize]-SortArray[1];
								END;
							(*
							{Variance and Std. Dev.}
								BEGIN
									Cell:=SortArray[13];
									Sum:=0;
									For i:=1 to SortSize DO
										BEGIN
											temp:=SortArray[i]-Cell;
											Sum:=Sum+temp*temp;
										END;
									Sum:=Sum div SortSize;
									if (myselection=9)
										THEN
											T1Ptr^[width-2]:=IntegerSqrt(Sum) 
										ELSE
											T1Ptr^[width-2]:=sum div 256;
								END;
							*)
							8:{Top Hat+}
								BEGIN
									ShellSort(SortArray,SortSize);
									cell:=SortArray[SortSize];
									SortArray[1]:=LocalArea[8]; {pull out center}
									SortArray[2]:=LocalArea[12];
									SortArray[3]:=LocalArea[13];
									SortArray[4]:=LocalArea[14];
									SortArray[5]:=LocalArea[18];
									SortSize:=5;
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-2]:=cell-SortArray[SortSize];
								END;
							9:{Top Hat-}
								BEGIN
									ShellSort(SortArray,SortSize);
									cell:=SortArray[1];
									SortArray[1]:=LocalArea[8]; {pull out center}
									SortArray[2]:=LocalArea[12];
									SortArray[3]:=LocalArea[13];
									SortArray[4]:=LocalArea[14];
									SortArray[5]:=LocalArea[18];
									SortSize:=5;
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-2]:=SortArray[1]-cell;
								END;
							END; {CASE}
						

						{x:=width-1;}
						
						{duplicate instead of rotate, col2 = col5}
						LocalArea[ 2]:=LocalArea[ 5];
						LocalArea[ 7]:=LocalArea[10];
						LocalArea[12]:=LocalArea[15];
						LocalArea[17]:=LocalArea[20];
						LocalArea[22]:=LocalArea[25];
						
						SortArray:=LocalArea;
						SortSize:=25;
						IF DoOctagonal THEN
							BEGIN {replace corners with values past 21}
								SortArray[1] :=SortArray[24];
								SortArray[5] :=SortArray[23];
								SortArray[21]:=SortArray[22];
								SortSize:=21;
							END;
						
						(*
						{alternative to sorting}
						IF (myselection=6) or (myselection=7) {not max only} THEN
							BEGIN
								min:=LocalArea[1];
								FOR zh:=2 TO SortSize DO
									IF min>LocalArea[zh] THEN min:=LocalArea[zh];
							END;
						
						IF (myselection=5) or (myselection=7) {not min only} THEN
							BEGIN
								max:=LocalArea[1];
								FOR zh:=2 TO SortSize DO
									IF max<LocalArea[zh] THEN max:=LocalArea[zh];
							END;
							
						CASE myselection OF
							5:T1Ptr^[width-1]:=max; 		{maximum}
							6:T1Ptr^[width-1]:=min;			{minimum}
							7:T1Ptr^[width-1]:=max-min; 	{range}
							END; {case}
						*)
						
						CASE myselection OF
							4:{median}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-1]:=SortArray[1+(SortSize div 2)];
								END;
							5:{Maximum}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-1]:=SortArray[SortSize];
								END;
							6:{Minimum}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-1]:=SortArray[1];
								END;
							7:{Range}
								BEGIN
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-1]:=SortArray[SortSize]-SortArray[1];
								END;
							(*
							{Variance and Std. Dev.}
								BEGIN
									Cell:=SortArray[13];
									Sum:=0;
									For i:=1 to SortSize DO
										BEGIN
											temp:=SortArray[i]-Cell;
											Sum:=Sum+temp*temp;
										END;
									Sum:=Sum div SortSize;
									if (myselection=9)
										THEN
											T1Ptr^[width-1]:=IntegerSqrt(Sum) 
										ELSE
											T1Ptr^[width-1]:=sum div 256;
								END;
							*)
							8:{Top Hat+}
								BEGIN
									ShellSort(SortArray,SortSize);
									cell:=SortArray[SortSize];
									SortArray[1]:=LocalArea[8]; {pull out center}
									SortArray[2]:=LocalArea[12];
									SortArray[3]:=LocalArea[13];
									SortArray[4]:=LocalArea[14];
									SortArray[5]:=LocalArea[18];
									SortSize:=5;
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-1]:=cell-SortArray[SortSize];
								END;
							9:{Top Hat-}
								BEGIN
									ShellSort(SortArray,SortSize);
									cell:=SortArray[1];
									SortArray[1]:=LocalArea[8]; {pull out center}
									SortArray[2]:=LocalArea[12];
									SortArray[3]:=LocalArea[13];
									SortArray[4]:=LocalArea[14];
									SortArray[5]:=LocalArea[18];
									SortSize:=5;
									ShellSort(SortArray,SortSize);
									T1Ptr^[width-1]:=SortArray[1]-cell;
								END;
							END; {CASE}
						
					END
				ELSE {color, 3 planes}
					BEGIN
						{what to do with a color image?}
					END;{else color}
				
				{update progress bar}
				ShowProgress(outRect.top,filterRect.bottom);
				
				{request next output portion}
				WITH outRect DO
					BEGIN
						top:=top+1;
						bottom:=bottom+1;
					END;
				
				{ask for new source rect}
				if outRect.bottom>filterrect.bottom then
					BEGIN
						SetRect(inRect,0,0,0,0); 		{flag that we're done}
						outrect:=inrect;
					END
				else
					BEGIN
						inrect.top:=outrect.top-(kFSize div 2);
						if inrect.top< filterrect.top then inrect.top:=filterrect.top;
						
						inrect.bottom:=outrect.bottom+(kFSize div 2);
						if inrect.bottom>filterrect.bottom then inrect.bottom:=filterrect.bottom;
					END;
				
			END;{with stuff^}
	END;{DoContinue}
	
	PROCEDURE DoFinish;
	BEGIN
		{this is used to throw away any temporary storage that may have been
		 allocated in DoStart.  called even if user cancels}
	END;{DoFinish}
	
	
VAR firstTime : Boolean; {quasi-global}

BEGIN {Dispatch - MAIN}
	{Main dispatching routine, Initialize and set up "global" variables,
	 perform operation specified by selector}
	 
	{I'm allowed to set up a data area, as a handle that I allocate the
	 first time through and gets disposed by Photoshop}
	firstTime:=data=NIL;
	
	if FirstTime then data:=NewHandle(sizeof(longint));
	
	gResult:=NoErr; {initialize the global error that
					 may accumulate in one on the sub-procedures}
	
	CASE selector OF
	filterSelectorAbout : 		DoAbout;
	filterSelectorParameters : 	DoParameters; {We're calling DoParameters in DoPrepare}
	filterSelectorPrepare : 	DoPrepare;
	filterSelectorStart : 		DoStart;
	filterSelectorContinue : 	DoContinue;
	filterSelectorFinish : 		DoFinish;
	END; {case}
	
	result:=gResult;
END; {Dispatch}

END. {unit}

(* NOTES

This drop-in provides a number of rank operations for monochrome images.
It ignores color images. The neigborhood is 5x5 pixels; by default, the
four corner pixels are skipped to produce an octagonal (and hence more
isotropic) result. Ranking of the 21 (or 25) pixel values is performed
using a Shell sort. The dialog allows the user to select several modes:
Median filter: a noise removal filter which saves the 11th (13th)
	brightest pixel in each neighborhood.
Maximum: saves the brightest pixel in each neighborhood. Often useful
	for removing dark noise or creating a background to isolate dark
	features.
Minimum: saves the darkest pixel in each neighborhood. Similar in use
	to the Maximum filter when the noise or features are bright.
Range: saves the difference between the brightest and darkest pixels.
	Useful to create an image showing texture or in some cases edges.
Top Hat (+): Performs two rank operations, one on the full 5x5 (square
	or octagonal) region and one on just the five central pixels
	forming a + in the center of the neighborhood, and saves the
	difference between the brightest pixels in each region. Isolates
	bright points and delineates edges.
Top Hat (-): Similar to Top Hat (+) but saves the difference between
	the darkest pixels in the large and small regions. Isolates dark
	points and delineates edges. Note that for edges, the Top Hat (+)
	and (-) methods will mark the inside and outside of edges depend-
	ing on the direction in which brightness varies across it.
After the filter has been run by selection from the "Other" submenu,
and the dialog selections made, repeating it by selecting Command-F
will use the same dialog selections. To change these, you must again
select Rank 5x5 from the submenu.

These filters are described in J. C. Russ "The Image Processing
Handbook" CRC Press, Boca Raton, FL (1992). The program is Copyright
 1992 by J. C. & J. C. Russ. It may be distributed at no charge.

*)